(in-package :contextl)

(defun function-name-p (name)
  (or (symbolp name)
      (and (consp name)
           (eq (car name) 'setf)
           (symbolp (cadr name))
           (null (cddr name)))))

(defun make-definer-name (name)
  (cond ((symbolp name)
         (intern (format nil "-~A-" (symbol-name name))
                 (symbol-package name)))
        ((consp name)
         (unless (function-name-p name)
           (error "Illegal function name: ~S." name))
         `(setf ,(intern (format nil "-~A-" (symbol-name (cadr name)))
                         (symbol-package (cadr name)))))
        (t (error "Illegal function name: ~S." name))))

(defmacro define-layered-function (name (&rest args) &body options)
  (let ((definer (or (let ((cons (assoc :definer options)))
                       (when cons
                         (when (cddr cons)
                           (error "Malformed :definer option: ~S." cons))
                         (let ((definer (cadr cons)))
                           (unless (function-name-p definer)
                             (error "Malformed definer: ~S." definer))
                           definer)))
                     (make-definer-name name))))
    (with-unique-names (layer-arg rest-arg)
      `(progn
         (defgeneric ,definer (,layer-arg ,@args)
           (:argument-precedence-order 
            ,@(let ((argument-precedence-order (assoc :argument-precedence-order options)))
                (if argument-precedence-order
                    (cdr argument-precedence-order)
                  (loop for arg in args
                        until (member arg lambda-list-keywords)
                        collect arg)))
            ,layer-arg)
           ,@(loop for option in options
                   unless (member (car option) '(:argument-precedence-order :definer))
                   collect option))
         (declaim (inline ,name))
         (defun ,name (&rest ,rest-arg)
           (declare (dynamic-extent ,rest-arg)
                    (optimize (speed 3) (debug 0) (safety 0)
                              (compilation-speed 0)))
           (apply #',definer (layer-context-prototype *active-context*) ,rest-arg))
         #',definer))))

(defun ensure-layered-function
       (name
        &rest initargs
        &key (lambda-list () lambda-list-p)
        (argument-precedence-order
         (loop for arg in lambda-list
               until (member arg lambda-list-keywords)
               collect arg))
        (definer (make-definer-name name))
        &allow-other-keys)
  (unless lambda-list-p
    (error "The layered function ~S must be initialized with a lambda list." name))
  (unless (function-name-p definer)
    (error "Illegal function name: ~S." definer))
  (let ((gf (let ((layer-arg (gensym "LAYER-ARG-")))
              (apply #'ensure-generic-function definer
                     :argument-precedence-order
                     `(,@argument-precedence-order ,layer-arg)
                     :lambda-list
                     `(,layer-arg ,@lambda-list)
                     initargs)))
        (rest-arg (gensym)))
    (setf (fdefinition name)
          (compile nil `(lambda (&rest ,rest-arg)
                          (declare (dynamic-extent ,rest-arg)
                                   (optimize (speed 3) (debug 0) (safety 0)
                                             (compilation-speed 0)))
                          (apply (the function ,gf)
                                 (layer-context-prototype *active-context*)
                                 ,rest-arg))))
    gf))
